local E = select(2, ...) -- Engine
local CO, UF = E:LoadModules("Config", "Unitframes")
local BA = CreateFrame("Frame", "CUI_AzeriteBarHolder", E.Parent)
BA.Autoload = true

local _
local XPBAR_TEXURE = [[Interface\AddOns\CUI\Textures\layout\modern\XPBar]]

function BA:InitEventHandler()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("AZERITE_ITEM_EXPERIENCE_CHANGED")
	self:RegisterEvent("PLAYER_XP_UPDATE")
	self:RegisterEvent("CVAR_UPDATE")
	
	self:SetScript("OnEvent", function(self, event, ...)
		if C_AzeriteItem.HasActiveAzeriteItem() then
			self:Show()
			self:UpdatePosition()
		else
			self:Hide()
		end
	end)
end

-- Basically shows the azerite bar as xp bar when the player is maxlevel (120)
function BA:UpdatePosition()
	if UnitLevel("player") ~= E.UNIT_MAXLEVEL then
		if self.CurrentPosition == "BOTTOM" then
			self:ClearAllPoints()
			self:SetPoint("BOTTOM", E.Parent, "BOTTOM", 0, 14)
			
			self:SetSize(749, 9)
		end
		
		self.CurrentPosition = "ABOVE"
	else
		if self.CurrentPosition == "ABOVE" then
			self:ClearAllPoints()
			self:SetPoint("BOTTOM", E.Parent, "BOTTOM", 0, 0)
			
			self:SetSize(749, 14)
		end
		
		self.CurrentPosition = "BOTTOM"
	end
end

function BA:__Construct()
	self:SetPoint("BOTTOM", E.Parent, "BOTTOM", 0, 14)
	self:SetSize(749, 9)
	self:SetParent(E.Parent)
	self:SetFrameStrata("MEDIUM")
	
	self.Bar = CreateFrame("Frame", "CUI_AzeriteBar", self, "AzeriteBarTemplate")
	
	self.Bar:ClearAllPoints()
	self.Bar:SetAllPoints(self)
	self.Bar:SetParent(self)
	self.Bar:Show()
	
	self.Bar.StatusBar:ClearAllPoints()
	self.Bar.StatusBar:SetAllPoints(self.Bar)
	
	self.Bar.StatusBar.ignoreFramePositionManager = true
	
	self.Bar.StatusBar:SetStatusBarTexture(XPBAR_TEXURE)
	
	self.Bar.StatusBar.Background:SetVertexColor(0.2,0.2,0.2,0.5)
	self.Bar.StatusBar.Underlay:SetVertexColor(0.2,0.2,0.2,0.9)
	
	self.Bar:SetFrameStrata("MEDIUM")
	self.Bar.StatusBar:SetFrameStrata("MEDIUM")
	
	self.Bar.OverlayFrame.Text:ClearAllPoints()
	self.Bar.OverlayFrame.Text:SetPoint("CENTER", self.Bar.OverlayFrame, "CENTER")
	
	UF:CreateBorder(self.Bar.StatusBar, nil, 1)
	self.Bar:ShowText() -- Always show text
	
	self.Bar:SetScript("OnEnter", function(self)
		self:UpdateOverlayFrameText()
		self:SetupPointsTooltip()
	end)
	self.Bar:SetScript("OnLeave", function(self)
		GameTooltip_Hide()
		self:CancelItemLoadCallback()
	end)
	
	
	self.CurrentPosition = "ABOVE"
	
	--E:CreateMover(self, "Azerite Bar")
end

-- Workaround for nasty error
function BA:IsTextLocked()
	return true
end

function BA:Init()
	self:__Construct()
	
	self:InitEventHandler()
end

E:AddModule("Bar_Azerite", BA)