local E = select(2, ...)
local CO, L, TT = E:LoadModules("Config", "Locale", "Tooltip")

local WM = CreateFrame("Frame")
WM.CurrentContinent, WM.CurrentZone, WM.MapInfo = nil, nil, nil
WM.Buttons = {}
WM.ButtonSize = 32

-- We use this to check if the zone has been changed. This is a workaround for the missing events.
-- Currently, the only way to check the changed zone is via onupdate.
WM.LastZone = -1


-- INCOMING_RESURRECT_CHANGED
-- [[Interface\RaidFrame\Raid-Icon-Rez]]

local WMM = nil

local EmptyMapInfo = {0, 0, 0, 0}
function WM:UpdateCurrentZone()
	-- WorldMapFrame:GetMapID()
	--/dump C_Map.GetWorldPosFromMapPos(WorldMapFrame:GetMapID(), CreateVector2D(0, 0))
	self.CurrentZone	= WorldMapFrame:GetMapID()
	--[[if not self.CurrentZone then
		self.MapInfo 		= EmptyMapInfo
	else
		self.MapInfo		= C_Map.GetMapInfo(self.CurrentZone)
	end]]
end

-- /dump C_Map.GetMapInfo(WorldMapFrame:GetMapID())
	-- /dump C_MapExplorationInfo.GetExploredAreaIDsAtPosition(WorldMapFrame:GetMapID(), CreateVector2D(10.00, 10.00))

-- The second return value of GetMapInfo() seems to be an accurate ID for the currently viewed zone
-- Fourth return value is a bool wether the current viewed map is a micro-dungeon aka. subzone
-- So we have to combine the MapContinent and MapZone with those values to receive an accurate view of what's displayed on the map.
-- /dump GetMapInfo()

-- Use GetPOITextureCoords(INDEX) somehow
function WM:SetEntranceTexture(object, instanceType, ...)
	-- local coords = { GetPOITextureCoords(1) }
	if type(...) ~= "string" then
		if instanceType == true then
			-- object:SetTexCoord(0.475, 0.42, 0.1275, 0.18)
			-- object:SetTexCoord(coords[1], coords[2], coords[3], coords[4])
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\raid]])
		else
			-- object:SetTexCoord(0.3725, 0.4275, 0.1275, 0.18)
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\dungeon]])
		end
	else
		if ... == "floorUp" then
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\door_up]])
		elseif ... == "floorDown" then
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\door_down]])
		elseif ... == "floorLeft" then
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\door_left]])
		elseif ... == "floorRight" then
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\door_right]])
		else
			object:SetTexture([[Interface\AddOns\CUI\Textures\icons\worldmap\door]])
		end
	end
end

function WM:CheckVisibility(object)
	if not object then return end
	
	-- If currently viewed map is a mini-dungeon
	if self.MapInfo[4] == true then
		-- If object is not in a mini-dungeon [sub-zone]
		if object.isInMiniDungeon == false then
			object:Hide()
		else
			-- If sub-zone floor matches
			if object.floor == self.MapInfo[2] then
				object:Show()
			else
				object:Hide()
			end
		end
	end
end

function WM:__Update(self, event, ...)
	self:UpdateCurrentZone()
	
	if not self.CurrentZone then return end
	if WM.LastZone == self.CurrentZone then return end
	
	self.mapWidth = WorldMapFrame.ScrollContainer.Child:GetWidth()
	self.mapHeight = WorldMapFrame.ScrollContainer.Child:GetHeight()
	
	E:debugprint("Zone: " .. self.CurrentZone)
	
	for k, v in pairs(WM.Buttons) do
		for _, f in pairs(v) do
			f:Hide()
		end
	end
	
	self.CurrentMarkerSet = nil
	
	if WMM.Markers[self.CurrentZone] then
		self.CurrentMarkerSet = WMM.Markers[self.CurrentZone]
		
		for k, instance in pairs(self.CurrentMarkerSet) do
			if type(instance) == "table" then
				if type(instance[1]) ~= "string" then
					self.currentInstanceInfo = { EJ_GetInstanceInfo(instance[1]) }
				end
				
				if (self.currentInstanceInfo and self.currentInstanceInfo[1]) or type(instance[1]) == "string" then
					if type(instance[1]) ~= "string" then
						self.currentInstanceName 	= self.currentInstanceInfo[1]
						-- self.currentInstanceDesc 	= self.currentInstanceInfo[2]
						self.currentInstanceX 		= instance[3]
						self.currentInstanceY 		= instance[4]
						self.currentInstanceNameOverride	= instance[5]
						self.navigateTarget 		= nil
					else
						self.currentInstanceName 	= instance[2]
						-- self.currentInstanceDesc 	= self.currentInstanceInfo[2]
						self.currentInstanceX 		= instance[3]
						self.currentInstanceY 		= instance[4]
						self.currentInstanceNameOverride	= nil
						self.navigateTarget			= instance[5]
					end
					
					if not self.Buttons[self.CurrentZone] then self.Buttons[self.CurrentZone] = {} end
					
					if self.Buttons[self.CurrentZone][k] then
						self.Buttons[self.CurrentZone][k]:Show()
						self.Buttons[self.CurrentZone][k].Icon:ClearAllPoints()
						self.Buttons[self.CurrentZone][k].Icon:SetPoint("CENTER", self.Buttons[self.CurrentZone][k], "CENTER")
					else
						self.Buttons[self.CurrentZone][k] = CreateFrame("Button")
						self.Buttons[self.CurrentZone][k]:SetSize(WM.ButtonSize, WM.ButtonSize)
						self.Buttons[self.CurrentZone][k]:SetPoint("TOPLEFT", WorldMapFrame.ScrollContainer.Child, "TOPLEFT", ((self.currentInstanceX / 100) * self.mapWidth) - (WM.ButtonSize / 2), -(((self.currentInstanceY / 100) * self.mapHeight) - (WM.ButtonSize / 2)))
						self.Buttons[self.CurrentZone][k]:SetParent(WorldMapFrame.ScrollContainer.Child)
						self.Buttons[self.CurrentZone][k]:SetFrameStrata("HIGH")
						
						if not self.currentInstanceNameOverride then
							self.Buttons[self.CurrentZone][k].instanceName = self.currentInstanceName
						else
							self.Buttons[self.CurrentZone][k].instanceName = string.format("%s: %s", self.currentInstanceName, self.currentInstanceNameOverride)
						end
						-- self.Buttons[self.CurrentZone][k].instanceDesc = self.currentInstanceDesc
						self.Buttons[self.CurrentZone][k].isRaid = instance[2]

						self.Buttons[self.CurrentZone][k]:EnableMouse(true)
						self.Buttons[self.CurrentZone][k]:SetScript("OnEnter", function(self)
							GameTooltip:SetOwner(self, "ANCHOR_TOP")
							GameTooltip:AddLine(self.instanceName)
							
							if self.isRaid == true then
								GameTooltip:AddLine(RAID)
							elseif self.isRaid == false then
								GameTooltip:AddLine(CALENDAR_TYPE_DUNGEON)
							end
							
							TT:UpdateStyle(nil)
						end)
						self.Buttons[self.CurrentZone][k]:SetScript("OnLeave", function(self)
							GameTooltip:Hide()
						end)
						
						if self.navigateTarget then
							
							self.Buttons[self.CurrentZone][k].navigateTarget = self.navigateTarget
							
							self.Buttons[self.CurrentZone][k]:SetScript("OnClick", function(self)
								WorldMapFrame:SetMapID(self.navigateTarget)
							end)
						end
						
						
						self.Buttons[self.CurrentZone][k].Icon = self.Buttons[self.CurrentZone][k]:CreateTexture(k, "OVERLAY")
						--self.Buttons[self.CurrentZone][k].Icon:SetAllPoints(self.Buttons[self.CurrentZone][k])
						self.Buttons[self.CurrentZone][k].Icon:SetPoint("CENTER", self.Buttons[self.CurrentZone][k], "CENTER")
						self.Buttons[self.CurrentZone][k].Icon:SetScale(1.1)
						
						self.Buttons[self.CurrentZone][k].Icon:SetParent(self.Buttons[self.CurrentZone][k])
						
							WM:SetEntranceTexture(self.Buttons[self.CurrentZone][k].Icon, instance[2], instance[1])
						
						self.Buttons[self.CurrentZone][k].Icon:SetTexCoord(0, 1, 0, 1)
						
						self.Buttons[self.CurrentZone][k]:SetHighlightTexture("Interface/WorldMap/UI-QuestPoi-NumberIcons");
						self.Buttons[self.CurrentZone][k]:GetHighlightTexture():SetTexCoord(0.625, 0.750, 0.875, 1);
					end
				end
			end
		end
	end
	
	WM.LastZone = self.CurrentZone
end

function WM:InitMarkers()
	self:UnregisterEvent("ADDON_LOADED")
	
	WMM = E:GetModule("WorldmapMarkers")
	
	
	hooksecurefunc(WorldMapFrame, "OnMapChanged", function() 
		WM:__Update(WM)
	end);
	--[[WorldMapFrame:HookScript("OnShow", function(self)
		WM:SetScript("OnUpdate", function(self, elapsed)
			WM:__Update(WM)
		end)
	end)
	WorldMapFrame:HookScript("OnHide", function(self)
		WM:SetScript("OnUpdate", nil)
	end)]]
	
	
	-- No longer works in BfA. Nothing event driven works there anymore
	--self:RegisterEvent("ZONE_CHANGED")
end

function WM:Init()
	self.db = CO.db.profile.worldmap
	
	-- We have to use this in BfA to retrieve the current map id
	-- /dump WorldMapFrame:GetMapID()
	
	if not IsAddOnLoaded("CUI_WorldmapMarkers") then
		self:RegisterEvent("ADDON_LOADED")
	else
		WM:InitMarkers()
	end
	self:SetScript("OnEvent", function(self, event, ...)
		if event == "ADDON_LOADED" then
			if ... == "CUI_WorldmapMarkers" then
				WM:InitMarkers()
			end
		
			return
		end
		self:__Update(self, event, ...)
	end)
end

E:AddModule("Worldmap", WM)