local E = select(2, ...) -- Engine
local CO, TB = E:LoadModules("Config", "Bar_Totem")
TB.Autoload = true

local CreateFrame	= CreateFrame
local GetTotemInfo 	= GetTotemInfo
local MAX_TOTEMS	= MAX_TOTEMS

TB.E				= CreateFrame("Frame") -- TB Event
local ButtonSize	= 40


local TotemExists, CDStart, CDDuration, Icon, Kids
function TB:__Update()
	
	for _, v in pairs(self.Bar.Buttons) do
		TotemExists, _, CDStart, CDDuration, Icon = GetTotemInfo(v.slot)
		
		if TotemExists then
			-- How to get rid of that nameless border?
			Kids = { v:GetChildren() };

			for _, child in ipairs(Kids) do
				child:Hide() -- First: Hide all childs
			end
			
			 -- Second: Show Icon and Cooldown again!
			v.Icon:Show()
			v.Cooldown:Show()
		end
	end
end

function TB:__Construct()
	self.Bar = CreateFrame("Frame", "CUI_TotemBar", E.Parent)
	self.Bar:SetSize(ButtonSize * MAX_TOTEMS, ButtonSize) -- Make this controllable via config somehow. Maybe smart-sizing
	
	self.Bar.Buttons = {}
	
	local CurrentButton, CurrentIcon, CurrentIconTexture, CurrentIconCooldown, CurrentIconDuration
	for i = 1, MAX_TOTEMS do
		CurrentButton = _G["TotemFrameTotem" .. i]
		
		if CurrentButton then
			CurrentIcon = _G["TotemFrameTotem" .. i .. "Icon"]
			CurrentIconDuration = _G["TotemFrameTotem" .. i .. "Duration"]
			CurrentIconTexture = _G["TotemFrameTotem" .. i .. "IconTexture"]
			CurrentIconCooldown = _G["TotemFrameTotem" .. i .. "IconCooldown"]
			
			CurrentButton.Icon = CurrentIcon
			CurrentButton.Icon.Texture = CurrentIconTexture
			
			CurrentButton.Cooldown = CurrentIconCooldown
			
			CurrentButton.Duration = CurrentIconDuration
			CurrentButton.Duration:ClearAllPoints()
			CurrentButton.Duration:SetPoint("TOP", CurrentButton, "BOTTOM", 0, 0)
			
			
			CurrentButton:SetSize(ButtonSize, ButtonSize)
			CurrentButton:ClearAllPoints()
			CurrentButton:SetParent(self.Bar)
			CurrentButton:SetPoint("LEFT", self.Bar, "LEFT", ButtonSize * (i - 1), 0)
			
			CurrentIcon:ClearAllPoints()
			CurrentIcon:SetAllPoints(CurrentButton)
			E:SkinButtonIcon(CurrentIconTexture, E:GetUnitClassColor("player"))
			
			self.Bar.Buttons[i] = CurrentButton -- Cache button for easier access
		end
	end
	
	E:CreateMover(self.Bar, "Totem-Bar")
	
	self.E:SetScript("OnEvent", function() self:__Update() end)
	self.E:RegisterEvent("PLAYER_ENTERING_WORLD")
	self.E:RegisterEvent("PLAYER_TOTEM_UPDATE")
	self:__Update() -- Initial Update
end

function TB:Init()
	-- self.db = CO.db.profile.totemBar
	
	self:__Construct()
end

E:AddModule("Bar_Totem", TB)