local E = select(2, ...) -- Engine
local L, UF = E:LoadModules("Locale", "Unitframes")
local AP = CreateFrame("Frame")
AP.Autoload = nil

AP.E = CreateFrame("Frame")

local select 					= select
local format 					= string.format
local pairs 					= pairs
local UnitClass 				= UnitClass
local GetSpecialization 		= GetSpecialization
local UnitHealthMax 			= UnitHealthMax
local UnitPower 				= UnitPower
local UnitPowerMax 				= UnitPowerMax
local UnitStagger 				= UnitStagger
local WarlockPowerBar_UnitPower = WarlockPowerBar_UnitPower


local AlternatePowers = {4,5,7,9,12,16} -- Supported (separate) power types








local function GetAltPowerValue(id)
	local value
	if id == 30 then
		value = UnitStagger("player")
	elseif id == 7 then
		value = WarlockPowerBar_UnitPower("player")
	else
		value = UnitPower("player", id)
	end

	return value
end

local function GetAltPowerMax(id)
	local value
	if id == 30 then
		-- Use 60% of maximum HP for stagger (ID = 30)
		value = UnitHealthMax("player") * 0.6
	else
		value = UnitPowerMax("player", id)
	end

	return value
end

-- @TODO: Refactor the alternate power system
function UF:SetupAlternatePower()

	UF.AlternatePowerUpdate = false
	if UF.LastAlternatePowerFrame then UF.LastAlternatePowerFrame:Hide() end

	local barName, altPower, altPowerOverlay, currentAltPowerCluster, AlternatePowerMainFrame, currentSpecialization, currentVisibilityCondition
	AlternatePowerMax = 0
	local gap = 5 -- 5 px gap between frames

	AlternatePowerId 		= nil
	CurrentClass 			= select(3, UnitClass("player"))
	currentSpecialization 	= GetSpecialization()

	for k,v in pairs(E.ClassPowers) do
		if k == CurrentClass then
			-- Force override to just show druid combo points when in feral form
			if k == 11 then
				AlternatePowerId = 4
				AlternatePowerMax = GetAltPowerMax(AlternatePowerId)
				currentVisibilityCondition = "[form:2] 1;0"

				break
			end
			if v[currentSpecialization] then
				if E:tableContainsValue(AlternatePowers, v[currentSpecialization]) or
					(CurrentClass == 7 and (currentSpecialization == 1 or currentSpecialization == 2)) or
					(CurrentClass == 5 and (currentSpecialization == 3)) then
					AlternatePowerId = v[currentSpecialization]
					AlternatePowerMax = GetAltPowerMax(AlternatePowerId)
					currentVisibilityCondition = "1"

					break -- We got what we wanted
				end
			end
		end
	end

	-- If no active alternate power was found
	if not AlternatePowerId then
		if UF.LastAlternatePowerFrame then
			UF.LastAlternatePowerFrame:Hide()
		end
		return
	else
		if E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. "Cluster"] then
			UF.LastAlternatePowerFrame = E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. "Cluster"]
			UF.LastAlternatePowerFrame:Show()
		end
	end

	if not E.Frames["AlternatePowerMainFrame"] then
		AlternatePowerMainFrame = CreateFrame("Frame", "AlternatePowerMainFrame", E.Parent, "SecureHandlerStateTemplate")
		AlternatePowerMainFrame:SetSize(295, 15)

			E.Frames["AlternatePowerMainFrame"] = AlternatePowerMainFrame

		AlternatePowerMainFrame:SetAttribute("_onstate-visible", [[
			if newstate == 1 then
				self:Show();
			else
				self:Hide();
			end
		]]);
		
		E:CreateMover(AlternatePowerMainFrame, L["alternatePower"])

		-- Post-load position, since we create this after the initial loading
		E:LoadMoverPositions(AlternatePowerMainFrame)
	else
		AlternatePowerMainFrame = E.Frames["AlternatePowerMainFrame"]
	end

	if not InCombatLockdown() then
		UnregisterStateDriver(AlternatePowerMainFrame, "visible")
		RegisterStateDriver(AlternatePowerMainFrame, "visible", currentVisibilityCondition)
	end

	-- OnUpdate runes
	if (select(3, UnitClass("player"))) == 6 then
		AlternatePowerMainFrame:SetScript("OnUpdate", function(self, elapsed) UF:UpdateAlternatePower() end)
	end

	-- Create just when it doesn't already exist!
	-- Also, those are the SEPARATED frames!
	if not E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. "1"] and E.PowerTypesDisplayType[AlternatePowerId] then
		--E:NewFrame(type, name, strata, sizeX, sizeY, point, parent, enablemouse, enablemousewheel, enablekeyboard, template)

		currentAltPowerCluster = E:NewFrame("Frame", "AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. "Cluster","LOW",295,100)
		UF.LastAlternatePowerFrame = currentAltPowerCluster

		currentAltPowerCluster:SetParent(AlternatePowerMainFrame)
		currentAltPowerCluster:ClearAllPoints()
		currentAltPowerCluster:SetPoint("CENTER", AlternatePowerMainFrame, "CENTER", 0, 0)

		for i=1,AlternatePowerMax do
			barName = "AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. i

			UF:CreateBar(barName, "MEDIUM", ((295 / AlternatePowerMax) - gap), 10, {"CENTER", currentAltPowerCluster, "CENTER"}, currentAltPowerCluster, 1, 0, 0)

			altPower = E.Frames[barName]
			altPowerOverlay = E.Frames[barName .. "Overlay"]

			altPowerOverlay:SetMinMaxValues(0,UF.ALTERNATE_MAX_VALUE)
			altPowerOverlay:SetValue(0)

			E.LibSmooth:SmoothBar(altPowerOverlay)

			--E:PushFrame(altPower, 0, 40)
			altPower:SetPoint("LEFT", currentAltPowerCluster, "LEFT", (((295 / AlternatePowerMax)*(i-1)) + gap / AlternatePowerMax), 0)
		end
	end
	-- Create just when it doesn't already exist!
	-- Also, those are the NON-SEPARATED frames!
	if not E.Frames["AlternatePower".. AlternatePowerId] and not E.PowerTypesDisplayType[AlternatePowerId] then
		--E:NewFrame(type, name, strata, sizeX, sizeY, point, parent, enablemouse, enablemousewheel, enablekeyboard, template)

		currentAltPowerCluster = E:NewFrame("Frame", "AlternatePower".. AlternatePowerId .. "-" .. "Cluster","LOW",295,100)
		UF.LastAlternatePowerFrame = currentAltPowerCluster

		currentAltPowerCluster:SetPoint("CENTER", AlternatePowerMainFrame, "CENTER", 0, 0)

			barName = "AlternatePower".. AlternatePowerId .. "-"

			UF:CreateBar(barName, "MEDIUM", 295 - gap, 10, {"CENTER", currentAltPowerCluster, "CENTER"}, currentAltPowerCluster, 1, 0, 0)

			altPower = E.Frames[barName]
			altPowerOverlay = E.Frames[barName .. "Overlay"]

			altPowerOverlay:SetMinMaxValues(0,AlternatePowerMax)
			altPowerOverlay:SetValue(0)

			E.LibSmooth:SmoothBar(altPowerOverlay)

			--E:PushFrame(altPower, 0, 40)
			altPower:SetPoint("CENTER", currentAltPowerCluster, "CENTER", 0, 0)
	end


	-- Since the frames were already created at this point, we can simply update it
	if AlternatePowerId then UF.AlternatePowerUpdate = true; UF:UpdateAlternatePower() end
end

function UF:UpdateAlternatePower()
	if not AlternatePowerId or (not E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. "Cluster"] and not E.Frames["AlternatePower".. AlternatePowerId .. "-" .. "Cluster"]) or not UF.AlternatePowerUpdate then return end
	local altPowerOverlay, altPowerValue, UpdateTime, updateValueTo, currentRuneCooldownStart, currentRuneCooldownDuration, currentRuneReady

	-- Get alt power value
	altPowerValue 	= GetAltPowerValue(AlternatePowerId)
	--if AlternatePowerId == 5 or altPowerValue == UF.LAST_ALT_POWER_VALUE then return end
	-- Iterate if the power is based on a separated display
	if E.PowerTypesDisplayType[AlternatePowerId] then
		for i=1,AlternatePowerMax do
			if not E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. i .."Overlay"] then return end

				if not AlternatePowerId then return end

				-- DEATHKNIGHT RUNES
				if AlternatePowerId == 5 then

					currentRuneCooldownStart, currentRuneCooldownDuration, currentRuneReady = GetRuneCooldown(i)

					-- Calculate the cooldown progress
					if not currentRuneReady then
						updateValueTo = (((currentRuneCooldownStart+currentRuneCooldownDuration-GetTime()) / currentRuneCooldownDuration) * UF.ALTERNATE_MAX_VALUE)
						if updateValueTo < 0 then
							updateValueTo = updateValueTo * -1
						end
					else
						updateValueTo = UF.ALTERNATE_MAX_VALUE
					end

				-- WARLOCK SHARDS
				elseif AlternatePowerId == 7 then
					SoulShards = WarlockPowerBar_UnitPower("player")

					local currentShard = i
					local currentShardValue = SoulShards - currentShard

					if currentShard <= SoulShards then
						updateValueTo = UF.ALTERNATE_MAX_VALUE
					else
						if currentShard - SoulShards > 0 and currentShard - SoulShards < 1 then
							updateValueTo = ((1-(currentShard - SoulShards)) * UF.ALTERNATE_MAX_VALUE)
						else
							updateValueTo = 0
						end
					end
				-- MONK STAGGER
				elseif AlternatePowerId == 30 then
					updateValueTo = (GetAltPowerMax(AlternatePowerId) / GetAltPowerValue(AlternatePowerId)) * UF.ALTERNATE_MAX_VALUE

				-- DEFAULT BEHAVIOUR
				else

					updateValueTo = UF.ALTERNATE_MAX_VALUE
				end

			-- Get current overlay frame
			altPowerOverlay	= E.Frames["AlternatePower".. AlternatePowerId .. "-" .. AlternatePowerMax .. "-" .. i .."Overlay"]
			if not altPowerOverlay then return end

			if updateValueTo == UF.ALTERNATE_MAX_VALUE then
				altPowerOverlay:GetStatusBarTexture():SetVertexColor(E.ClassColors[CurrentClass][1],E.ClassColors[CurrentClass][2],E.ClassColors[CurrentClass][3],1)
			else
				-- Perform color update to be sure
				altPowerOverlay:GetStatusBarTexture():SetVertexColor(E.PowerColors[5][1],E.PowerColors[5][2],E.PowerColors[5][3],E.PowerColors[5][4])
			end


				-- Use new power value since it is VERY likely to have already changed!
				if (i <= UnitPower("player", AlternatePowerId)) or AlternatePowerId == 5 or AlternatePowerId == 7 or AlternatePowerId == 30 then
					-- If i is less than altPowerValue then we expect the frames to be filled
					if updateValueTo < 0 then updateValueTo = updateValueTo * -1 end
					altPowerOverlay:SetValue(updateValueTo)
				else
					-- If i is greater than altPowerValue then we have reached the empty frames
					altPowerOverlay:SetValue(0)
				end

		end
	else
		-- Get current overlay frame
		altPowerOverlay 	= E.Frames["AlternatePower".. AlternatePowerId .. "-" .. "Overlay"]
		-- Perform color update to be sure
		altPowerOverlay:GetStatusBarTexture():SetVertexColor(E.PowerColors[AlternatePowerId][1],E.PowerColors[AlternatePowerId][2],E.PowerColors[AlternatePowerId][3],E.PowerColors[AlternatePowerId][4])
		altPowerOverlay:SetValue(GetAltPowerValue(AlternatePowerId))
	end

	UF.LAST_ALT_POWER_VALUE = altPowerValue
end




function AP:CreateHolder()
	self.Holder = CreateFrame("Frame", "CUI_AlternatePower", E.Parent, "SecureHandlerStateTemplate")
	self.Holder:SetPoint("CENTER", E.Parent, "CENTER")
	self.Holder:SetSize(215, 12)
	
	self:UpdateCurrentPowerInfo()
	
	E:SetVisibilityHandler(self.Holder)
	self:UpdateHolderVisibility()
	
	E:CreateMover(self.Holder, L["alternatePower"])
end

function AP:UpdateHolderVisibility()
	if not InCombatLockdown() then
		UnregisterStateDriver(self.Holder, "visible")
		RegisterStateDriver(self.Holder, "visible", self.VisibilityCondition)
	end
end

function AP:OnlyShowCurrent()
	for _, v in pairs(self.StoredPowers) do
		if v and v.Holder then
			v.Holder:Hide()
			print(v.Holder:GetName())
		end
	end
	
	self.CurrentPower.Holder:Show()
end

function AP:RepositionSegments()
	local SizeX = self.Holder:GetWidth() / self.PowerMax
	local SizeY = self.Holder:GetHeight()
	
	for i = 1, self.PowerMax do
		self.CurrentPower.Bars[i]:ClearAllPoints()
		self.CurrentPower.Bars[i]:SetPoint("LEFT", self.Holder, "LEFT", SizeX * (i - 1) + ((i - 1) * (5)), 0)
		self.CurrentPower.Bars[i]:SetSize(SizeX, SizeY)
	end
end

function AP:CreateSegment(i, SizeX, SizeY)
	self.CurrentPower.Bars[i] = UF:CreateBar(format("CUI_AlternatePowerSegment%d", i), "LOW", SizeX, SizeY, {"LEFT", self.Holder, "LEFT", SizeX * (i - 1) + ((i - 1) * (5)), 0}, self.Holder, 0, 0, 0, nil)
	self.CurrentPower.Bars[i].SetMinMaxValues(0, 100)
end

function AP:CreateSegments()

	if not self.CurrentPower.Bars then self.CurrentPower.Bars = {} end

	local SizeX, SizeY
	SizeX = self.Holder:GetWidth() / self.PowerMax
	SizeY = self.Holder:GetHeight()
	
	for i = 1, self.PowerMax do
		if not (self.CurrentPower.Bars[i]) then
			self:CreateSegment(i, SizeX, SizeY)	
		end
		
		if i > self.NumMaxSegments then self.NumMaxSegments = i end
	end
end

function AP:UpdateSegments()
	self:UpdateCurrentPowerInfo()
	
	if self.NumMaxSegments < self.PowerMax then
		self:CreateSegments()
	end
	
	
	-- Hide all that are not needed currently
	for i = 1, self.NumMaxSegments do
		if i <= self.PowerMax then
			self.CurrentPower.Bars[i]:Show()
		else
			self.CurrentPower.Bars[i]:Hide()
		end
	end
	
	self:RepositionSegments()
end

function AP:UpdateValue()
	if not self.CurrentPower then return end
	self.PowerValue = GetAltPowerValue(self.PowerId)
	
	for i = 1, self.PowerMax do
		if self.CurrentPower.Bars[i] then
			if i <= self.PowerValue then
				self.CurrentPower.Bars[i].SetValue(100)
			else
				self.CurrentPower.Bars[i].SetValue(0)
			end
		end
	end
end

-- Write all neccessary data
function AP:UpdateCurrentPowerInfo()
	self.PlayerClass	= select(3, UnitClass("player"))
	self.PlayerSpec 	= GetSpecialization()
	
	for k,v in pairs(E.ClassPowers) do
		if k == self.PlayerClass then
			-- Force override to just show druid combo points when in feral form
			if k == 11 then
				self.PowerId = 4
				self.PowerMax = GetAltPowerMax(self.PowerId)
				self.VisibilityCondition = "[form:2] 1;0"

				break
			end
			if v[self.PlayerSpec] then
				if E:tableContainsValue(AlternatePowers, v[self.PlayerSpec]) or
					(self.PlayerClass == 7 and (self.PlayerSpec == 1 or self.PlayerSpec == 2)) or
					(self.PlayerClass == 5 and (self.PlayerSpec == 3)) then
					
						self.PowerId = v[self.PlayerSpec]
						self.PowerMax = GetAltPowerMax(self.PowerId)
						self.VisibilityCondition = "1"

						break -- We got what we wanted
				end
			end
		end
	end
end

local function __OnEvent(self, event, ...)
	if select(3, UnitClass("player")) ~= 6 then
		if event == "UNIT_POWER_UPDATE" or event == "RUNE_POWER_UPDATE" then self:UpdateValue() end
	end

	if event == "UNIT_DISPLAYPOWER" or event == "PLAYER_SPECIALIZATION_CHANGED" or event == "SPELLS_CHANGED" then self:UpdateSegments() end
end

function AP:InitEventHandler()
	self:RegisterUnitEvent("UNIT_POWER_UPDATE", "player")
	self:RegisterUnitEvent("UNIT_DISPLAYPOWER", "player")
	self:RegisterEvent("RUNE_POWER_UPDATE")
	self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self:RegisterEvent("SPELLS_CHANGED")
	
	self:SetScript("OnEvent", __OnEvent)
end

function AP:__Construct()
	-- Here we store the information about the current power displayed
	self.CurrentPower = {} 
	self.StoredPowers = {}
	
	self.NumMaxSegments = -1
	
	self:CreateHolder()
	self:UpdateSegments()
	self:InitEventHandler()
end

function AP:Init()
	self:__Construct()
end

E:AddModule("Alternate_Power", AP)