local E = select(2, ...) -- Engine
local CO,L,UF,BA,UAUR,TT = E:LoadModules("Config", "Locale", "Unitframes", "Bar_Auras", "UnitAuras", "Tooltip")

--[[-------------------------------------------------------------------------

	We are caching globals, since making those functions local,
	results in a slight performance boost and therefore
	in less CPU time. Exactly what we are aiming for.

-------------------------------------------------------------------------]]--
local _
local format 					= string.format
local wipe						= wipe
local tremove 					= table.remove
local tsort 					= table.sort
local CreateFrame 				= CreateFrame
local DebuffTypeColor 			= DebuffTypeColor
local UnitExists 				= UnitExists
local UnitCanAttack 			= UnitCanAttack
local UnitAura 					= UnitAura
-----------------------------------------------------------------------------

BA.E = CreateFrame("Frame")
BA.Bars = {}
BA.Auras = {}

local BAR_NUM, BAR_GAP_X, BAR_GAP_Y, BAR_SIZE_X, BAR_SIZE_Y = 15, 0, 2, 295, 18


BA.BAR_NUM = BAR_NUM

local function SortByExpiration(a,b)
	if a and b then
		if a[6] < b[6] then
			return true
		elseif a[6] > b[6] then
			return false
		end
	end
end

function BA:UpdateHeader(Header)
	local SizeX, SizeY = 0, 0
	
	for i=1,BAR_NUM do
		SizeX = Header[i]:GetWidth() + CO.db.profile.auras.units[Header.Unit].aurabars.iconSize
		SizeY = SizeY + Header[i]:GetHeight() + CO.db.profile.auras.units[Header.Unit].aurabars.gapY
	end
	
	Header:SetSize(SizeX, SizeY)
	E:UpdateMoverDimensions(Header)
end

function BA:ToggleBars(Unit)
	if BA.Bars[Unit].ForceShow then
		BA.Bars[Unit].ForceShow = nil
	else
		BA.Bars[Unit].ForceShow = true
	end
end

function BA:UpdateName(Object, Aura)
	if Aura[3] and Aura[3] > 1 then
		Object:SetText(format("%s [%s]", Aura[1], Aura[3]))
	else
		if Object:GetText() ~= Aura[1] then
			Object:SetText(Aura[1])
		end
	end
end

function BA:UpdateTime(Object, TimeLeft)
	if TimeLeft > 10 then self.UpdateTimePlaces = 0; else self.UpdateTimePlaces = 1; end
	
	Object:SetText(E:FormatTime(TimeLeft, self.UpdateTimePlaces))
end

function BA:UpdateTexture(Object, Texture)
	if Object.CurrentTexture ~= Texture then
		Object:SetTexture(Texture)
		Object.CurrentTexture = Texture
	end
end

function BA:UpdateBarValues(Object, TimeLeft, Duration)
	Object:SetValue(TimeLeft)
	if Object.CurrentDuration ~= Duration then
		Object:SetMinMaxValues(0, Duration)
		Object.CurrentDuration = Duration
	end
end

-- In this method, we make heavy use of control variables, since those have the least impact on memory
function BA:UpdateBarColor(Object, RGB)
	
	if Object.Bar.Overlay.RGB ~= RGB then
		Object.Bar.Overlay:GetStatusBarTexture():SetVertexColor(RGB.r, RGB.g, RGB.b, 1)
		
		Object.Bar.Overlay.RGB = RGB
	end
	if Object.Bar.Border.RGB ~= RGB then
		Object.Bar.Border:SetBackdropBorderColor(RGB.r, RGB.g, RGB.b, 1)
		
		Object.Bar.Border.RGB = RGB
		
		
		if Object.Icon.Tex.RGB ~= RGB then
			E:SkinButtonIcon(Object.Icon.Tex, RGB)
			
			Object.Icon.Tex.RGB = RGB
		end
	end
end

local TooltipUnit, TooltipRealIndex, TooltipParent
function BA:BuildTooltip(self)
	TooltipParent = self:GetParent()
	TooltipUnit = TooltipParent:GetParent().Unit
	if not BA.Auras[TooltipUnit] or not BA.Auras[TooltipUnit][TooltipParent.Index] then return end
	
	-- We have to retrieve the real aura index, since we do remove and sort auras from the table
	for k, v in pairs(BA.Auras[TooltipUnit]) do
		-- Because of the way how we assign the tables, we can do a direct comparison
		if v == BA.Auras[TooltipUnit][TooltipParent.Index] then
			TooltipRealIndex = v.RealIndex
		end
	end	
	
	if TooltipRealIndex then
		GameTooltip:SetOwner(self, "ANCHOR_TOP")
		GameTooltip:SetUnitAura(TooltipUnit, TooltipRealIndex, BA.Auras[TooltipUnit].AuraType)
		
		GameTooltip:Show()
	end
end

local HideAllFrame
function BA:HideAll(Unit)
	HideAllFrame = self.Bars[Unit]
		for i=1, BAR_NUM do
			if not HideAllFrame[i].IsHidden then
				HideAllFrame[i]:Hide()
				HideAllFrame[i].IsHidden = true
			end
		end
		
	HideAllFrame.AllHidden = true
end

function BA:CreateBars(Unit)
	local BarColor
	
	self.Bars[Unit] = CreateFrame("Frame", format("AuraBarContainer%s", Unit))
	self.Bars[Unit]:SetPoint(E.STR.CENTER, E.Parent, E.STR.CENTER)
	self.Bars[Unit]:SetSize(BAR_SIZE_X, BAR_SIZE_Y * BAR_NUM)
	
	self.Bars[Unit].Unit = Unit
	
	E:CreateMover(self.Bars[Unit], format("%s %s", L[Unit], L["AuraBars"]), E.STR.BOTTOMLEFT)
	
	for i=1,BAR_NUM do
		self.Bars[Unit][i] = CreateFrame("Frame", format("AuraBar%s%s", Unit, i)) -- Acts as a parent
		self.Bars[Unit][i]:SetPoint(E.STR.BOTTOMLEFT, self.Bars[Unit], E.STR.BOTTOMLEFT, BAR_GAP_X, (CO.db.profile.auras.units[Unit].aurabars.gapY + BAR_SIZE_Y) * (i - 1))
		self.Bars[Unit][i]:SetSize(BAR_SIZE_X, BAR_SIZE_Y)
		self.Bars[Unit][i]:SetParent(self.Bars[Unit])
		
		self:CreateIcon(self.Bars[Unit][i], format("AuraBar%s%sIcon", Unit, i))
		self.Bars[Unit][i].Icon:SetScript("OnEnter", function(self) BA:BuildTooltip(self) end)
		self.Bars[Unit][i].Icon:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
		
		self.Bars[Unit][i].Bar = UF:CreateBar(format("AuraBar%s%sOverlay", Unit, i), "LOW", BAR_SIZE_X - BAR_SIZE_Y, BAR_SIZE_Y, {"LEFT", self.Bars[Unit][i].Icon, "RIGHT", 0, 0}, self.Bars[Unit][i].Icon, false, false, false)
		self.Bars[Unit][i].Bar:SetParent(self.Bars[Unit][i].Icon)
		
		-- We don't need this here
		E.LibSmooth:ResetBar(self.Bars[Unit][i].Bar.Overlay)
		
		BarColor = E:GetUnitReactionColor(Unit)
		self.Bars[Unit][i].Bar.Overlay:GetStatusBarTexture():SetVertexColor(BarColor.r, BarColor.g, BarColor.b, 1)
		
		self:InitFonts(self.Bars[Unit][i].Bar.Overlay)
		
		self.Bars[Unit][i].Index = i
		
		self.Bars[Unit][i]:Hide()
	end
	
	-- Post script to prevent issues
	self.Bars[Unit]:SetScript("OnUpdate", function(self, elapsed)
		
		if self.ForceShow then
			for i=1, BAR_NUM do
				self[i]:Show()
				self[i].Icon.Tex:SetTexture(134400)
				self[i].Bar.Overlay.name:SetText(format("Placeholder Aura %s", i))
				self[i].Bar.Overlay.time:SetText(E:Round((i / BAR_NUM) * BAR_NUM, 2) .. "s")
				BA:UpdateBarValues(self[i].Bar.Overlay, i, BAR_NUM)
			end
			return
		end
		
		if not (BA.Auras[Unit] and BA.Auras[Unit][1]) and not self.AllHidden then BA:HideAll(Unit); return end
		
		-- Execute post-sort when needed
		if not self.UpdateInProgress and self.QueueAuraSort then
			tsort(BA.Auras[Unit], SortByExpiration)
			
			self.QueueAuraSort = nil
		end
		
		
		for i=1, BAR_NUM do
			-- If we don't have any auras of the unit, return
			if UnitExists(self.Unit) and BA.Auras[Unit] and BA.Auras[Unit][i] and BA.Auras[Unit][i][5] > 0 then
				
				self.CurrentBar = self[i]
				self.CurrentAura = BA.Auras[Unit][i]
				
				if self.CurrentAura then
					self.CurrentBar.duration = self.CurrentAura[5]
					self.CurrentBar.timeLeft = self.CurrentAura[6] - GetTime()
					
					if self.CurrentBar.timeLeft < 0 then self.CurrentAura[i] = nil else
						
						BA:UpdateBarValues(BA.Bars[Unit][i].Bar.Overlay, self.CurrentBar.timeLeft, self.CurrentBar.duration)
						
						-- Determine color
						self.AuraColor = UAUR:GetAuraColor(self.CurrentAura[4], Unit, BA.Auras[Unit].AuraType)
						
						-- Perform updates
						BA:UpdateBarColor(self.CurrentBar, self.AuraColor)
						BA:UpdateTexture(self.CurrentBar.Icon.Tex, self.CurrentAura[2])
						BA:UpdateName(self.CurrentBar.Bar.Overlay.name, self.CurrentAura)
						BA:UpdateTime(self.CurrentBar.Bar.Overlay.time, self.CurrentBar.timeLeft)
					end
				end
				if not self.CurrentBar:IsVisible() then self.CurrentBar:Show(); self.CurrentBar.IsHidden = nil; self.AllHidden = nil end
			else
				if self[i]:IsVisible() then self[i]:Hide() end
			end
		end
	end)
	
	BA:UpdateAuras(Unit)
end

function BA:CreateIcon(F, Name)
	F.Icon = CreateFrame("Frame", Name)
	F.Icon:SetPoint(E.STR.LEFT, F, E.STR.LEFT)
	F.Icon:SetSize(BAR_SIZE_Y, BAR_SIZE_Y)
	F.Icon:SetParent(F)
	
	F.Icon:EnableMouse(true)
	
	F.Icon.Tex = F.Icon:CreateTexture(nil, "OVERLAY")
	F.Icon.Tex:SetAllPoints(F.Icon)
	
	E:SkinButtonIcon(F.Icon.Tex)
end

function BA:InitFonts(F)
	local FontType = "FRIZQT__.TTF"
	local Fonts = {["time"] = {"RIGHT", 100, 18, -5}, ["name"] = {"LEFT", 150, 18, 5}} -- Alignment, Width, Height, XOffset
	
	for n,v in pairs(Fonts) do
		F[n] = F:CreateFontString(nil, "ARTWORK")
		E:InitializeFontFrame(F[n], "ARTWORK", font, 11, {1,0.96,0.41}, 1, {0,0}, "", v[2], v[3], F, v[1], {1,1})
		F[n]:ClearAllPoints()
		F[n]:SetParent(F)
		F[n]:SetJustifyH(v[1])
		F[n]:SetPoint(v[1], F, v[1], v[4], 0)
	end
end

local AuraName, AuraSubName, AuraTexture, AuraCount, AuraDType, AuraDuration, AuraExpirationTime, UnitAuraClass, CurrentAuraIndex
function BA:UpdateAuras(Unit)
	if not UnitExists(Unit) or not self.Bars[Unit] then return end
	
	-- We use this to gain more control over the sort process, since it sometimesseems to run when we update the aura cache
	self.UpdateInProgress = true
	CurrentAuraIndex = 1
	
	-- Start with a clean table
	if self.Auras[Unit] then wipe(self.Auras[Unit]) end
	if not self.Auras[Unit] then self.Auras[Unit] = {} end
	
	if UnitCanAttack(Unit, E.STR.player) then UnitAuraClass = E.STR.HARMFUL; else UnitAuraClass = E.STR.HELPFUL; end
	
	-- Iterate until we reach the last auraID of the unit
	while true do
		
		AuraName, AuraTexture, AuraCount, AuraDType, AuraDuration, AuraExpirationTime = UnitAura(Unit, CurrentAuraIndex, UnitAuraClass .. "|PLAYER")
		if not AuraName then
			if self.Auras[Unit] and self.Auras[Unit][CurrentAuraIndex] then
				self.Auras[Unit][CurrentAuraIndex] = nil
			end
			
			-- If aura would have been the first one. R.I.P
			if CurrentAuraIndex == 1 then
				wipe(self.Auras[Unit])
			end
			
				break
		end
		
		if not self.Auras[Unit][CurrentAuraIndex] then self.Auras[Unit][CurrentAuraIndex] = {} end
		
		
		-- Used for tooltips
		self.Auras[Unit][CurrentAuraIndex].RealIndex = CurrentAuraIndex
		
		-- Numerical indexing for better sort results (results at all)
		self.Auras[Unit][CurrentAuraIndex][1] = AuraName
		self.Auras[Unit][CurrentAuraIndex][2] = AuraTexture
		self.Auras[Unit][CurrentAuraIndex][3] = AuraCount
		self.Auras[Unit][CurrentAuraIndex][4] = AuraDType
		self.Auras[Unit][CurrentAuraIndex][5] = AuraDuration
		self.Auras[Unit][CurrentAuraIndex][6] = AuraExpirationTime
		
		-- Also cache the aura type [Harmful or Helpful]
		self.Auras[Unit].AuraType = UnitAuraClass
		
		CurrentAuraIndex = CurrentAuraIndex + 1
	end
	
	self.UpdateInProgress = nil
	
	-- Remove uneccessary entries. Start from the end. Otherwise the loop would cancel after removing the first entry
	for i=#self.Auras[Unit],1,-1 do
		-- Remove auras with a duration higher than 5 minutes or no duration at all
		-- @TODO: Make this highly customizable and migrate this method to an external file inside the AUR namespace
		-- So we can use it for more than just the aurabars
		if self.Auras[Unit][i] and (self.Auras[Unit][i][5] == 0 or self.Auras[Unit][i][5] >= 300) then
			tremove(self.Auras[Unit], i)
		end
	end
	
	-- Sort by expiration time
	if self.Auras[Unit] and not self.UpdateInProgress then
		tsort(self.Auras[Unit], SortByExpiration)
	elseif self.UpdateInProgress then
		self.QueueAuraSort = true
	end
end

function BA:Init()
	CO = E:GetModule("Config")
	self.db = CO.db.profile.auras
	
	self.E:RegisterEvent("UNIT_AURA", "player", "target")
	self.E:RegisterEvent("PLAYER_TARGET_CHANGED")
	self.E:SetScript("OnEvent", function(selfE, event, ...)
		if event == "PLAYER_TARGET_CHANGED" then self:UpdateAuras("target") else
			self:UpdateAuras(...)
		end
	end)
	
	self:CreateBars("player")
	self:CreateBars("target")
end

E:AddModule("Bar_Auras", BA)