local E = select(2, ...) -- Engine
local BC,CO,L,UF,AB,LT,AUR,TT = E:LoadModules("Bar_Cast", "Config", "Locale", "Unitframes", "Actionbars", "Layout", "Auras", "Tooltip")

local _, ClassColor
	BC.Bars = {}
	
local pairs							= pairs
local select						= select
local format						= string.format
local GetNetStats					= GetNetStats
local GetTime						= GetTime
local CastingBarFrame_ApplyAlpha	= CastingBarFrame_ApplyAlpha
local INTERRUPTED					= INTERRUPTED

-- string.format usage:
-- string.format("CUI_%sCastbar%s", unit, index)
-- %s can be used an unlimited amount of times. Also do additional args
local BarBaseName 			= "CUI_%sCastbar%s"
local IconSize 				= 23
local interruptible 		= {0.85, 0.85, 0.85}
local notInterruptible 		= {1, 0.772, 0.380}

function BC:Register(B, BName)
	BC.Bars[BName] = B
end

function BC:Get(BName)
	return BC.Bars[BName]
end

function BC:GetIndex(unit)
	local i = 1
	for k, v in pairs(BC.Bars) do
		if v.unit == unit then
			i = i + 1
		end
	end
	
	return i
end

function BC:AddText(b, n, a, x, y)
	b[n] = b.Overlay:CreateFontString(nil, "ARTWORK")
	local f = b[n]
	
	E:InitializeFontFrame(b[n], "ARTWORK", CO.Media:Fetch("font", self.db["fontType"]), 11, {0.8,0.8,0.8}, 1, {0,0}, "", 0, 0, b.Overlay, "RIGHT", {1,1})
	f:SetFont(CO.Media:Fetch("font", self.db["fontType"]), 11, "")
	f:ClearAllPoints()
	f:SetPoint(a, b, a)
	f:SetJustifyH(a)
	f:SetJustifyV("MIDDLE")
	
	if x and y then E:PushFrame(f, x, y) end
end

function BC:ToggleMovers(s)
	for k, v in pairs(BC.Bars) do
		BC.Bars[k].moverEnabled = s
		
		if s == true then
			BC.Bars[k]:Show()
			BC.Bars[k]:SetAlpha(1)
		else
			BC.Bars[k]:Hide()
		end
	end
end

function BC:AddLagBar(b)
	b.LagBar = CreateFrame("Frame", nil, b.Overlay)
	--b.LagBar:SetAllPoints(b.Overlay)
	b.LagBar:SetSize(b.Overlay:GetWidth(), b.Overlay:GetHeight())
	b.LagBar:SetPoint("RIGHT", b.Overlay, "RIGHT")
	b.LagBarTex = b.LagBar:CreateTexture(nil)
	b.LagBarTex:SetAllPoints(b.LagBar)
	b.LagBarTex:SetPoint("RIGHT", b.LagBar)
	b.LagBarTex:SetColorTexture(0.65, 0, 0, 0.5)
	
	BC:UpdateLagBar(b, false)
end

function BC:UpdateLagBar(b, s)
	if s == false then
		b.LagBar:Hide()
		return
	else
		b.LagBar:Show()
	end
	
	local timePerPixel, lagBarWidth
	local LagWorld = select(4, GetNetStats())
	local min, max = b.GetMinMaxValues()
	-- We are always assuming the min max values are timings
	-- We use the delta value to determine the real needed width of the LagBar
	local delta = max - min
	timePerPixel = b.Overlay:GetWidth() / delta
	
	lagBarWidth = (LagWorld / 1000) * timePerPixel

	b.LagBar:SetWidth(lagBarWidth)
end

function BC:AddSpark(b)
	b.Spark = b.Overlay:CreateTexture(nil, "OVERLAY")
	b.Spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
	b.Spark:SetBlendMode("ADD")
	b.Spark:SetSize(32, 55)
end

function BC:SetSparkPosition(o, isChanneling)
	if isChanneling then
		o.sparkPositionX = ((o.value / o.maxValue) * o.Overlay:GetWidth()) + 2
	else
		o.sparkPositionX = ((o.currentTime / o.endTime) * o.Overlay:GetWidth()) - 2
	end
	
	o.Spark:SetPoint("CENTER", o, "LEFT", o.sparkPositionX, 0)
end

function BC:AddEventHandler(bar)

	-- Register a bunch (all) of spellcast events (all we need)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_START", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_STOP", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_FAILED", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_SUCCEEDED", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_INTERRUPTED", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_INTERRUPTIBLE", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_DELAYED", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_START", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_STOP", bar.Unit)
	bar:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_UPDATE", bar.Unit)
	bar:RegisterEvent("UNIT_TARGET")
	if bar.Unit == "target" then
		bar:RegisterEvent("PLAYER_TARGET_CHANGED")
	elseif bar.Unit == "focus" then
		bar:RegisterEvent("PLAYER_FOCUS_CHANGED")
	end
	
	-- Interruptor
	bar:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	
	bar:SetScript("OnEvent", function(self, event, ...)
		-- Interruptor handler START
		----------------------------
		if self:IsVisible() and event == "COMBAT_LOG_EVENT_UNFILTERED" then
			-- Probably the most efficient way we can go
			
			_, self.combatLogInfoType, _, _, self.combatLogInfoName, _, _, self.combatLogInfoGUID = CombatLogGetCurrentEventInfo()
			
			if self.combatLogInfoType == "SPELL_INTERRUPT" and self.combatLogInfoGUID == UnitGUID(self.Unit) then
				self.Interruptor = self.combatLogInfoName
				
				-- Check if the interruptor name is valid. Environmental effects like quaking leave this at nil, a.e.
				if self.Interruptor then
					self.Name:SetText(format("%s [%s]", INTERRUPTED, self.Interruptor))
				end
			end
			
			-- End call directly, since we do not want the script to iterate through everything else. This does get fired REALLY rapidly in combat.
			return
		end
		-- Interruptor handler END
		----------------------------
		
		-- print(event)
		
		self.eventUnit = ...
		
		-- Hide when there is no such unit
		if not UnitExists(bar.Unit) then self:Hide(); return; end
		if self.eventUnit ~= bar.Unit and not event == "PLAYER_TARGET_CHANGED" then return end
		
		if (event == "PLAYER_TARGET_CHANGED" or event == "PLAYER_FOCUS_CHANGED") and not (UnitCastingInfo(bar.Unit) or UnitChannelInfo(bar.Unit)) then
			self.casting = nil
			self.channeling = nil
			self.fadeOut = true
			self.holdTime = 0
			
			self:Hide()
		end
		
		if event == "UNIT_SPELLCAST_START" or ((event == "PLAYER_TARGET_CHANGED" or event == "PLAYER_FOCUS_CHANGED") and UnitCastingInfo(bar.Unit)) then
			self.SpellName, _, self.SpellTexture, self.startTime, self.endTime, SpellIsTradeSkill, SpellCastID, self.notInterruptible  = UnitCastingInfo(bar.Unit)
			if self.SpellName then
				CastingBarFrame_ApplyAlpha(self, 1.0)
				self.holdTime = 0
				self.casting = true
				self.channeling = nil
				self.fadeOut = nil
				
				self.endTime = (self.endTime - self.startTime) / 1000
				self.currentTime = (GetTime() - (self.startTime / 1000))
				self.SetMinMaxValues(0, self.endTime)
				self.SetValue(self.currentTime)
				
				if ( self.Spark ) then
					self.Spark:Show()
				end
				self.Name:SetText(self.SpellName)
				self.Icon.Tex:SetTexture(self.SpellTexture)
				
				
				if self.notInterruptible then
					bar.SetOverlayColor(notInterruptible[1], notInterruptible[2], notInterruptible[3], 0.95)
					E:SkinButtonIcon(self.Icon.Tex, notInterruptible)
				else
					bar.SetOverlayColor(interruptible[1], interruptible[2], interruptible[3], 0.95)
					E:SkinButtonIcon(self.Icon.Tex, interruptible)
				end
				
				-- Keep lagbar enabled for non-player units, when the cast is interruptible to help with interrupting it on high latency
				if bar.Unit ~= "player" and self.notInterruptible then BC:UpdateLagBar(self, false) else BC:UpdateLagBar(self, true) end
				
				self:Show()
				self:SetAlpha(1)
				
				-- self.holdTime = GetTime() + CASTING_BAR_HOLD_TIME
			else
				self.casting = nil
				self.channeling = nil
				self.fadeOut = true
				self.holdTime = 0
				
				self:Hide()
			end
			
		
			
		elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
			if self.Name:GetText() == INTERRUPTED or self.channeling then return end
			
			-- If still casting [Fix for passive auras that also trigger this event]
			if UnitCastingInfo(bar.Unit) then return end
			
			self.casting = nil
			self.channeling = nil
			self.fadeOut = true
			self.holdTime = CASTING_BAR_HOLD_TIME
			
			if ( self.Spark ) then
				self.Spark:Hide()
			end
			self.SetValue(select(2, self.Overlay:GetMinMaxValues()))
			self.Time:SetText("")
			
			bar.SetOverlayColor(0,0.85,0, 0.95)
			--local name, rank, displayName, texture, startTime, endTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo(bar.Unit)
			--if not name then
								
			--end

		elseif event == "UNIT_SPELLCAST_CHANNEL_STOP" then
			-- If still casting [Fix for passive auras that also trigger this event]
			if not self.channeling or UnitChannelInfo(bar.Unit) then return end
			
			self.casting = nil
			self.channeling = nil
			self.fadeOut = true
			self.holdTime = CASTING_BAR_HOLD_TIME
			
			if ( self.Spark ) then
				self.Spark:Hide()
			end
			self.SetValue(select(2, self.Overlay:GetMinMaxValues()))
			self.Time:SetText("")
			
			bar.SetOverlayColor(0,0.85,0, 0.95)
		
		elseif event == "UNIT_SPELLCAST_FAILED" then
			if UnitCastingInfo(bar.Unit) then return end
			
			if not self.casting then return end
				self.casting = nil
				self.channeling = nil
				self.fadeOut = true
				self.holdTime = 0
				
				self:Hide()
			
		elseif event == "UNIT_SPELLCAST_FAILED_QUIET" then
			
		elseif event == "UNIT_SPELLCAST_INTERRUPTED" then
			if not self.casting then return end
			-- If still casting [Fix for passive auras that also trigger this event]
			--if UnitCastingInfo(bar.Unit) then return end
			
			if ( self.Spark ) then
				self.Spark:Hide()
			end
			self.SetValue(select(2, self.Overlay:GetMinMaxValues()))
			self.Time:SetText("")
			
			-- We set the interruptor name in the COMBAT_LOG_EVENT_UNFILTERED handler
			self.Name:SetText(INTERRUPTED)
			
			bar.SetOverlayColor(0.85,0,0, 0.95)
		
			self.casting = nil
			self.channeling = nil
			self.fadeOut = true
			self.holdTime = GetTime() + CASTING_BAR_HOLD_TIME
		elseif event == "UNIT_SPELLCAST_INTERRUPTIBLE" then
			
		elseif event == "UNIT_SPELLCAST_DELAYED" then
			if ( self:IsShown() ) then
				self.SpellName, _, _, self.startTime, self.endTime, _, _, _  = UnitCastingInfo(bar.Unit)
				if not self.SpellName then
					-- if there is no name, there is no bar
					self:Hide();
					return;
				end
				
				self.endTime = (self.endTime - self.startTime) / 1000
				self.currentTime = (GetTime() - (self.startTime / 1000))
				self.SetMinMaxValues(0, self.endTime)
			end
			
		-- Immediate interruption (Spellcast failed directly)
		elseif event == "UNIT_SPELLCAST_STOP" then
			if not self.casting then return end
			
			self.casting = nil
			self.channeling = nil
			self.fadeOut = true
			self.holdTime = 0
			
			self:Hide()
		
		elseif event == "UNIT_SPELLCAST_CHANNEL_START" or ((event == "PLAYER_TARGET_CHANGED" or event == "PLAYER_FOCUS_CHANGED") and UnitChannelInfo(bar.Unit)) then
			self.SpellName, _, self.SpellTexture, self.startTime, self.endTime, self.SpellIsTradeSkill, _, _ = UnitChannelInfo(bar.Unit)
			
			if not self.SpellName then
				-- if there is no name, there is no bar
				self:Hide()
				return
			end

			bar.SetOverlayColor(ClassColor[1],ClassColor[2],ClassColor[3], 0.95)
				
			self:Show()
			self:SetAlpha(1)

			self.Name:SetText(self.SpellName)
			self.Icon.Tex:SetTexture(self.SpellTexture)
			E:SkinButtonIcon(self.Icon.Tex, ClassColor)
			
			self.maxValue = (self.endTime - self.startTime) / 1000
			self.value = (self.endTime / 1000) - GetTime()
			self.minValue = 0
			self.SetMinMaxValues(0, self.maxValue)
			self.SetValue(self.value)
			self.casting = nil
			self.channeling = true
			
			if ( self.Spark ) then
				self.Spark:Show()
			end
			
			BC:UpdateLagBar(self, false)
			
			self.holdTime = GetTime() + CASTING_BAR_HOLD_TIME
			
		-- Channel delay
		elseif event == "UNIT_SPELLCAST_CHANNEL_UPDATE" then
			if ( self:IsShown() ) then
				self.SpellName, _, self.SpellTexture, self.startTime, self.endTime, self.SpellIsTradeSkill, _, _ = UnitChannelInfo(bar.Unit)
				if not self.SpellName then
					-- if there is no name, there is no bar
					self:Hide();
					return;
				end

				self.value = ((self.endTime / 1000) - GetTime());
				self.maxValue = (self.endTime - self.startTime) / 1000;
				self.SetMinMaxValues(0, self.maxValue);
				self.SetValue(self.value);
			end
			
		end
	end)
	
	bar.OnUpdate = self.OnUpdate
	bar:HookScript("OnUpdate", bar.OnUpdate)
end

function BC:OnUpdate(elapsed)
	if self.moverEnabled then return end
	
	if ( self.casting ) then
		
		self.currentTime = self.currentTime + elapsed;
		if ( self.currentTime >= self.endTime ) then
			self.SetValue(self.endTime);
			return;
		end
		self.SetValue(self.currentTime)
		self.Time:SetText(E:Round(self.endTime - self.currentTime, 1))
		if ( self.Flash ) then
			self.Flash:Hide();
		end
		if ( self.Spark ) then
			BC:SetSparkPosition(self, false)
		end
		
	elseif ( self.channeling ) then
		self.value = self.value - elapsed;
		if ( self.value <= self.minValue ) then
			-- CastingBarFrame_FinishSpell(self, self.Spark, self.Flash);
			return;
		end
		
		self.SetValue(self.value);
		self.Time:SetText(E:Round(self.value, 1))
		if ( self.Flash ) then
			self.Flash:Hide();
		end
		if ( self.Spark ) then
			BC:SetSparkPosition(self, true)
		end
	elseif ( GetTime() < self.holdTime ) then
		return;
	elseif ( self.flash ) then
		local alpha = 0;
		if ( self.Flash ) then
			alpha = self.Flash:GetAlpha() + CASTING_BAR_FLASH_STEP;
		end
		if ( alpha < 1 ) then
			if ( self.Flash ) then
				self.Flash:SetAlpha(alpha);
			end
		else
			if ( self.Flash ) then
				self.Flash:SetAlpha(1.0);
			end
			self.flash = nil;
		end
	elseif ( self.fadeOut ) then
		local alpha = self:GetAlpha() - 0.015;
		if ( alpha > 0 ) then
			-- CastingBarFrame_ApplyAlpha(self, alpha);
			self:SetAlpha(alpha)
		else
			self.fadeOut = nil;
			self:Hide();
		end
	end
end

function BC:Create(unit)
	local i = self:GetIndex(unit) -- Get index
	local name = format("CUI_%sCastbar%s", unit, i)
	
	local Bar = UF:CreateBar(name, "LOW", 235, 25, {"CENTER", E.Parent, "CENTER"}, E.Parent)
	E.LibSmooth:ResetBar(Bar.Overlay) -- Leaving the smooth anim on somehow causes the bar to not go at a 100%. This results in the LagBar simply being useless and just looks weird
	Bar.Background.SetRGBA(nil, nil, nil, 0.95)
	
	Bar.Unit = unit
	
	Bar.Icon = CreateFrame("Frame", "CUI_CastbarIconHolder", Bar)
	Bar.Icon:SetSize(IconSize, IconSize)
	Bar.Icon:SetPoint("LEFT", Bar, "LEFT", -IconSize, 0)
	
	Bar.Icon.Tex = Bar.Icon:CreateTexture(nil, "OVERLAY")
	Bar.Icon.Tex:SetTexCoord(0.06, 0.94, 0.06, 0.94)
	Bar.Icon.Tex:SetParent(Bar.Icon)
	Bar.Icon.Tex:SetAllPoints(Bar.Icon)
	
	self:AddText(Bar, "Time", "RIGHT", -10, 0)
	self:AddText(Bar, "Name", "LEFT", 5, 0)
	
	-- LAG BAR
	self:AddLagBar(Bar)
	-- SPARK
	self:AddSpark(Bar)
	-- Functionality
	self:AddEventHandler(Bar)
	-- Initial Hide
	Bar:Hide()
	-- Add mover
	E:CreateMover(Bar, format("%s-%s%s", L[Bar.Unit], L["castbar"], i))
	-- Register castbar to engine
	self:Register(Bar, name)
end

function BC:Init()
	CO = E:GetModule("Config")
	UF = E:LoadModules("Unitframes")
	self.db = CO.db.profile.castbar
	self.moverdb = CO.db.profile.movers
	
	ClassColor = E:GetUnitClassColor("player")
	
	BC:Create("player")
	BC:Create("target")
	BC:Create("focus")
end

E:AddModule("Bar_Cast", BC)