local E = select(2, ...) -- Engine
local CO, UF, AUR, UAUR = E:LoadModules("Config", "Unitframes", "Auras", "UnitAuras")

local _
local CreateFrame		= CreateFrame
local format			= string.format
local pairs				= pairs
local UnitAura			= UnitAura

UAUR.E = CreateFrame("Frame")
UAUR.Auras = {}
UAUR.Holders = {}

UAUR.AURA_SIZE 				= 32 -- X and Y size
UAUR.AURA_MARGIN 			= 5 -- Margin between auras in px
UAUR.BUFF_PER_ROW 			= 8
UAUR.BUFF_NUM_PLAYER 		= 32
UAUR.DEBUFF_NUM_PLAYER 		= 16
UAUR.AURA_TYPES 			= {"HELPFUL", "HARMFUL"}
local HarmfulColor			= {0.85, 0, 0}

--[[---------------------------------------------

	Aura Table:
		UAUR.Auras
			-> [Unit]
				-> [AuraClass](HELPFUL or HARMFUL)
					-> [AuraIndex]
						-> [1] = AuraName
						-> [2] = AuraTexture
						-> [3] = AuraCount
						-> [4] = AuraDType
						-> [5] = AuraDuration
						-> [6] = AuraExpirationTime
				-> UnitAuraClass = HELPFUL or HARMFUL, based on faction to player

------------------------------------------------]]

function UAUR:CreateIcon(Slot)
	Slot.Tex = Slot:CreateTexture(nil, "BACKGROUND")
	Slot.Tex:SetParent(Slot)
	Slot.Tex:SetAllPoints(Slot)
	Slot.Tex:SetSize(32, 32)
	
	Slot.Highlight = Slot:CreateTexture(nil, "HIGHLIGHT")
	Slot.Highlight:SetColorTexture(1, 1, 1, 0.45)
	
	E:SkinButtonIcon(Slot.Tex, {0.3, 0.3, 0.8, 1})
	
	self:SetupTooltip(Slot)
	self:SetupCooldown(Slot)
	
	self:InitFonts(Slot, Slot.Cooldown)
end

function UAUR:SetupCooldown(Slot)
	Slot.Cooldown = CreateFrame("Cooldown", nil, Slot, "CooldownFrameTemplate")
	Slot.Cooldown:SetParent(Slot)
	Slot.Cooldown:SetAllPoints(Slot)
	--Slot.Cooldown:SetFrameStrata("TOOLTIP")
	Slot.Cooldown:SetReverse(true)
	Slot.Cooldown:Show()
end

local function AuraButton_OnEnter(self)
	GameTooltip:SetOwner(self)
	GameTooltip:SetUnitAura(self:GetParent():GetParent().Unit, self.RealIndex, self.AuraClass)
end

function UAUR:SetupTooltip(Slot)
	Slot:SetScript("OnEnter", AuraButton_OnEnter)
	Slot:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
end

local ConfigLoaderFrame, SlotStartPosition
local function ConfigLoader(UnitOrFrame)
	
	if type(UnitOrFrame) == "string" then
		ConfigLoaderFrame = UF.Frames[UnitOrFrame]
	elseif type(UnitOrFrame) == "table" then
		ConfigLoaderFrame = UnitOrFrame
	end
	if ConfigLoaderFrame and not ConfigLoaderFrame.AuraHolder or not ConfigLoaderFrame then return end
	
	
	if ConfigLoaderFrame.EnableAuras == false then
		ConfigLoaderFrame.AuraHolder:SetScript("OnUpdate", nil)
		ConfigLoaderFrame.AuraHolder:Hide()
	else
		UAUR:SetUpdateHandler(ConfigLoaderFrame)
		
		ConfigLoaderFrame.AuraHolder:ClearAllPoints()
		ConfigLoaderFrame.AuraHolder:SetPoint(ConfigLoaderFrame.AuraPosition, ConfigLoaderFrame, ConfigLoaderFrame.AuraRelativePosition, ConfigLoaderFrame.AuraOffsetX, ConfigLoaderFrame.AuraOffsetY)
		ConfigLoaderFrame.AuraHolder:SetSize(ConfigLoaderFrame:GetWidth(), ConfigLoaderFrame.AuraSize * 1)
		
		
		-- Smart positioning within the holder
		SlotStartPosition = ConfigLoaderFrame.AuraGrowthDirectionY == "UP" and "BOTTOM" or "TOP"
		--SlotStartPosition = ConfigLoaderFrame.AuraGrowthDirectionX == "LEFT" and SlotStartPosition .. "RIGHT" or SlotStartPosition .. "LEFT"
		if ConfigLoaderFrame.AuraGrowthDirectionX == "LEFT" then
			SlotStartPosition = SlotStartPosition .. "RIGHT"
		elseif ConfigLoaderFrame.AuraGrowthDirectionX == "RIGHT" then
			SlotStartPosition = SlotStartPosition .. "LEFT"
		elseif ConfigLoaderFrame.AuraGrowthDirectionX == "CENTER" then
			SlotStartPosition = "CENTER"
		end
		
		
		for i = 1, ConfigLoaderFrame.AuraSlots do
			-- @TODO: Add an smart handler that manages the position.
			-- @TODO: Add gaps between each slot
			
			if not ConfigLoaderFrame.AuraHolder.AuraSlot[i] then
				UAUR:BuildAuras(ConfigLoaderFrame)
			end
			
			ConfigLoaderFrame.AuraHolder.AuraSlot[i]:ClearAllPoints()
			ConfigLoaderFrame.AuraHolder.AuraSlot[i]:SetPoint(SlotStartPosition, ConfigLoaderFrame.AuraHolder, SlotStartPosition, UAUR:GetGrowthDirection(ConfigLoaderFrame.AuraSize, ConfigLoaderFrame.AuraGrowthDirectionX, i, ConfigLoaderFrame.AuraSlots) * (i-1), 0)
			ConfigLoaderFrame.AuraHolder.AuraSlot[i]:SetSize(ConfigLoaderFrame.AuraSize, ConfigLoaderFrame.AuraSize)
		end
		
		-- Check for more than necessary active slots
			for i=1, 20 do
				if ConfigLoaderFrame.AuraHolder.AuraSlot[i] and (i > ConfigLoaderFrame.AuraSlots) then
					ConfigLoaderFrame.AuraHolder.AuraSlot[i]:Hide()
				end
			end
			
			ConfigLoaderFrame.TypeStart = ConfigLoaderFrame.DebuffsFirst and 2 or 1
			ConfigLoaderFrame.TypeEnd 	= ConfigLoaderFrame.DebuffsFirst and 1 or 2
			ConfigLoaderFrame.TypeStep 	= ConfigLoaderFrame.DebuffsFirst and -1 or 1
			
		ConfigLoaderFrame.EndAtSlot 		= ConfigLoaderFrame.AuraSlots
		ConfigLoaderFrame.updateTimer		= 0
		ConfigLoaderFrame.updateInterval	= 0.035
		
		UAUR:UpdateIcons(ConfigLoaderFrame.Unit)
	end
end

-- Maybe hook into the general UF update method?
function UAUR:LoadConfig(Unit)
	UF:PerformForUnits(Unit, ConfigLoader)
end

function UAUR:HolderEvent(Frame)
	self:UpdateAuras(Frame.Unit)
	self:SetUpdateHandler(Frame)
end

function UAUR:SetHolderEvent(Frame)
	Frame.AuraHolder:RegisterUnitEvent("PLAYER_ENTERING_WORLD")
	Frame.AuraHolder:RegisterUnitEvent("UNIT_AURA", Frame.Unit)
	
	Frame.AuraHolder:SetScript("OnEvent", function(holder, event, ...)
		UAUR:HolderEvent(Frame)
	end)
end

local CurrentAuraIndex, CurrentAura, CurrentTimeLeft
function UAUR:SetUpdateHandler(F)
	
	-- If no aura exists, remove handler
	if (not self.Auras[F.Unit] or (self.Auras[Unit] and not (self.Auras[F.Unit][self.AURA_TYPES[1]] and self.Auras[F.Unit][self.AURA_TYPES[2]])))
		or not F.EnableAuras then
		F:SetScript("OnUpdate", nil)
		F.AuraHolder:Hide()
		
		for i=1, F.AuraSlots do
			if F.AuraHolder.AuraSlot[i] then
				F.AuraHolder.AuraSlot[i]:Hide()
			end
		end
		
		return
	else
		F.AuraHolder:Show()
	end
	
	-- Prevent from setting it up more than once
	if F:GetScript("OnUpdate") then return end
	
	F:SetScript("OnUpdate", function(handler, elapsed)
		-- We have to loop through every aura slot.
		-- We can end the loop as soon as the corresponding aura does not exist
		-- We ONLY do the necessary things here, that really have to be updated in real-time
		
			handler.updateTimer = handler.updateTimer + elapsed
		if handler.updateTimer < handler.updateInterval then return end
			handler.updateTimer = 0
		
		handler.BeginAtSlot 	= 1
		handler.CurrentSlotNum 	= 0
		
		-- Loop through both aura classes
		-- (I have the feling that this will crush our performance)		
		for i = handler.TypeStart, handler.TypeEnd, handler.TypeStep do
			handler.CurrentAuraType = self.AURA_TYPES[i]
			
			if UAUR:IsCorrectAuraType(handler) then
				if self.Auras[handler.Unit] and self.Auras[handler.Unit][handler.CurrentAuraType] then
					CurrentAuraIndex = 1
					
					-- Iterate through each slot
					for n = handler.BeginAtSlot, handler.EndAtSlot do
						handler.CurrentSlot = handler.AuraHolder.AuraSlot[n]
						CurrentAura = self.Auras[handler.Unit][handler.CurrentAuraType][CurrentAuraIndex]
						
						
						if CurrentAura and CurrentAura[1] then
							
							--CurrentAura[2] -- AuraName
							--CurrentAura[1] -- Texture
							--CurrentAura[2] -- AuraCount
							--CurrentAura[3] -- DType
							--CurrentAura[4] -- Duration
							--CurrentAura[5] -- Expiration
							
							-- Used for tooltips
							handler.CurrentSlot.RealIndex 	= CurrentAura.RealIndex
							handler.CurrentSlot.AuraClass 	= CurrentAura.AuraClass
							
							CurrentTimeLeft = CurrentAura[5] - GetTime()
							
							if CurrentTimeLeft > 0 and CurrentTimeLeft < 15 then
								handler.CurrentSlot.Time:SetText(E:FormatTime(CurrentTimeLeft, 1))
							elseif CurrentTimeLeft > 15 then
								handler.CurrentSlot.Time:SetText(E:FormatTime(CurrentTimeLeft, 0))
							else
								if handler.CurrentSlot.Time:GetText() ~= "" then
									handler.CurrentSlot.Time:SetText("")
								end
							end
							
							handler.CurrentSlotNum = handler.CurrentSlotNum + 1
						else
							handler.BeginAtSlot = handler.CurrentSlotNum + 1
							break
						end
						
						CurrentAuraIndex = CurrentAuraIndex + 1
					end
				end
			end
		end
	end)
end

local CurrentVisible
function UAUR:UpdateCenterPositioning(Holder)
	CurrentVisible = 0
	for i = 1, Holder.AuraSlots do
		if Holder.AuraHolder.AuraSlot[i]:IsVisible() then
			CurrentVisible = CurrentVisible + 1
		end
	end
	
	for i = 1, Holder.AuraSlots do
		Holder.AuraHolder.AuraSlot[i]:ClearAllPoints()
		Holder.AuraHolder.AuraSlot[i]:SetPoint("CENTER", Holder.AuraHolder, "CENTER", UAUR:GetGrowthDirection(Holder.AuraSize, "CENTER", i, CurrentVisible), 0)
	end
end

-- @TODO: Center requires active positioning
local GrowthDirection, GapX
function UAUR:GetGrowthDirection(Size, Direction, NumCurrent, NumVisible)
	GapX = 1
	
	if Direction == "UP" or Direction == "RIGHT" then
		GrowthDirection = Size
	elseif Direction == "DOWN" or Direction == "LEFT" then
		GrowthDirection = Size * (-1)
	elseif Direction == "CENTER" and NumCurrent and NumVisible then
		GrowthDirection = (NumCurrent - NumVisible / 2) * (Size + GapX) - Size / 2
	end
	
	return GrowthDirection
end

function UAUR:InitFonts(Slot, Parent)
	local FontType = "FRIZQT__.TTF"
	local Fonts = {["Time"] = {"CENTER", 0, 0, 0, 0, "MIDDLE"}, ["Count"] = {"BOTTOMRIGHT", 0, 0, -2, 2, "RIGHT"}} -- Alignment, Width, Height, XOffset
	
	for n,v in pairs(Fonts) do
		Parent = Parent or Slot
		
		Slot[n] = Slot:CreateFontString(nil, "ARTWORK")
		E:InitializeFontFrame(Slot[n], "ARTWORK", FontType, 11, {1,0.96,0.41}, 1, {0,0}, "", v[2], v[3], Parent, v[1], {0, 0}, "OUTLINE")
		Slot[n]:ClearAllPoints()
		Slot[n]:SetParent(Parent)
		Slot[n]:SetJustifyH(v[6])
		Slot[n]:SetPoint(v[1], Parent, v[1], v[4], v[5])
	end
end

function UAUR:RegisterHolder(Frame, Unit)
	if not self.Holders[Unit] then
		self.Holders[Unit] = {}
	end
	
	table.insert(self.Holders[Unit], Frame)
end

local BuildAurasHasHolder
function UAUR:BuildAuras(Frame)
	-- Create an aura holder frame as overlay
	-- This holder has a specific amount of slots and can display buffs as well as debuffs
	-- The frame has to bring some properties that tell this method how this slot should be created
	-- Whenever an aura is filled in, we set the onupdate handler for this aura
	-- When it no longer exists, we remove the handler
	-- We also want to make the holder completely movable across the screen.
	-- Means, we need an option to set the parent. Best would be from a predefined list
	
	BuildAurasHasHolder = nil
	
	-- Update instead of building when it already exists
	if Frame.AuraHolder then self:SetUpdateHandler(Frame); BuildAurasHasHolder = true; else		
		-- Profile unit holds the unit + index for raid40. Name is not really needed, but why not
		Frame.AuraHolder = CreateFrame("Frame", format("%sAuraHolder", Frame.ProfileUnit), Frame)
		Frame.AuraHolder:SetFrameStrata("LOW")
		Frame.AuraHolder.AuraSlot = {}
		Frame.UpdateAuras = function() self:HolderEvent(Frame) end
		
		self:RegisterHolder(Frame, Frame.Unit)
	end
	
	-- Build new slots of we need them
	for i = 1, Frame.AuraSlots do
		if not Frame.AuraHolder.AuraSlot[i] then
			Frame.AuraHolder.AuraSlot[i] = CreateFrame("Frame", format("CUI_AuraIcon%s", i), Frame.AuraHolder)
			
			Frame.AuraHolder.AuraSlot[i]:EnableMouse(true)
			
			self:CreateIcon(Frame.AuraHolder.AuraSlot[i])
		end
	end
	
	-- Init auras for this unit
	if not self.Auras[Frame.Unit] then self.Auras[Frame.Unit] = {} end
	for _, AuraClass in pairs(self.AURA_TYPES) do
		if not self.Auras[Frame.Unit][AuraClass] then self.Auras[Frame.Unit][AuraClass] = {} end
		
		for i = 1, 20 do
			self.Auras[Frame.Unit][AuraClass][i] = {}
		end
	end
	
	if not BuildAurasHasHolder then
		self:SetHolderEvent(Frame)
		self:SetUpdateHandler(Frame)
		
		self:LoadConfig(Frame) -- Initial update
	end
end

-- Copy of the method for Bar_Auras, since there seems to be no way of getting rid of the table reference without causing a lot of trouble
local AuraName, AuraSubName, AuraTexture, AuraCount, AuraDType, AuraDuration, AuraExpirationTime, UnitAuraClass, CurrentAuraIndex
function UAUR:UpdateAuraCache(Unit)
	-- Start with a clean table
	--if self.Auras[Unit] then wipe(self.Auras[Unit]) end
	if not self.Auras[Unit] then return end
	
	self.Auras[Unit].UnitAuraClass = UnitCanAttack(Unit, E.STR.player) and E.STR.HARMFUL or E.STR.HELPFUL
	
	for _, AuraClass in pairs(self.AURA_TYPES) do
		CurrentAuraIndex = 1
		
		-- Iterate until we reach the last auraID of the unit
		while true do
			
			if CurrentAuraIndex > 20 then break end
			
			AuraName, AuraTexture, AuraCount, AuraDType, AuraDuration, AuraExpirationTime = UnitAura(Unit, CurrentAuraIndex, AuraClass)
			
			--[[-------------------------------------------
				We now just write into the table
				Even when there are just nil values.
				This will avoid creation of garbage piles.
				
				The lack of checking if there is an aura
				also should speed things up a bit (Less CPU time)
			----------------------------------------------]]
			
			-- Used for tooltips
			self.Auras[Unit][AuraClass][CurrentAuraIndex].RealIndex = CurrentAuraIndex
			self.Auras[Unit][AuraClass][CurrentAuraIndex].AuraClass = AuraClass
			
			-- Numerical indexing for better sort results (results at all)
			--self.Auras[Unit][AuraClass][CurrentAuraIndex][1] = AuraName
			self.Auras[Unit][AuraClass][CurrentAuraIndex][1] = AuraTexture
			self.Auras[Unit][AuraClass][CurrentAuraIndex][2] = AuraCount
			self.Auras[Unit][AuraClass][CurrentAuraIndex][3] = AuraDType
			self.Auras[Unit][AuraClass][CurrentAuraIndex][4] = AuraDuration
			self.Auras[Unit][AuraClass][CurrentAuraIndex][5] = AuraExpirationTime
			
			if not AuraName then
				break
			end
			
			CurrentAuraIndex = CurrentAuraIndex + 1
		end
	end
end

local AuraTypeCurrent
function UAUR:IsCorrectAuraType(handler)

	if UnitCanAttack(handler.Unit, E.STR.player) then
		AuraTypeCurrent = handler.EnemyAuraOnlyFromType
	else
		AuraTypeCurrent = handler.FriendlyAuraOnlyFromType
	end
	
	if not AuraTypeCurrent then return true end
	if AuraTypeCurrent == handler.CurrentAuraType then return true else return false end
end

function UAUR:UpdateIcons(Unit)
	-- We have to loop through every aura slot.
		-- We can end the loop as soon as the corresponding aura does not exist
		-- We ONLY do the necessary things here, that really have to be updated in real-time
		for _, handler in pairs(self.Holders[Unit]) do
			
			handler.BeginAtSlot 	= 1
			handler.CurrentSlotNum 	= 0
			
			if handler.TypeStart and handler.TypeEnd and handler.TypeStep then
				
				-- Loop through both aura classes
				for i = handler.TypeStart, handler.TypeEnd, handler.TypeStep do
					handler.CurrentAuraType = self.AURA_TYPES[i]
					
					if self:IsCorrectAuraType(handler) then
					--
						if self.Auras[handler.Unit] and self.Auras[handler.Unit][handler.CurrentAuraType] then
							CurrentAuraIndex = 1
							
							-- Iterate through each slot
							for n = handler.BeginAtSlot, handler.EndAtSlot do
								handler.CurrentSlot = handler.AuraHolder.AuraSlot[n]
								CurrentAura = self.Auras[handler.Unit][handler.CurrentAuraType][CurrentAuraIndex]
								
								
								if CurrentAura and CurrentAura[1] then
									
									--CurrentAura[2] -- AuraName
									--CurrentAura[1] -- Texture
									--CurrentAura[2] -- AuraCount
									--CurrentAura[3] -- DType
									--CurrentAura[4] -- Duration
									--CurrentAura[5] -- Expiration
									
									-- Used for tooltips
									handler.CurrentSlot.RealIndex 	= CurrentAura.RealIndex
									handler.CurrentSlot.AuraClass 	= CurrentAura.AuraClass
									
									
									if CurrentAura[2] > 1 then
										handler.CurrentSlot.Count:SetText(CurrentAura[2])
									else
										handler.CurrentSlot.Count:SetText("")
									end
									
									handler.CurrentSlot.Tex:SetTexture(CurrentAura[1])
									E:SkinButtonIcon(handler.CurrentSlot.Tex, self:GetAuraColor(CurrentAura[3], handler.Unit, handler.CurrentAuraType))
									
									if handler.CurrentSlot.Cooldown and CurrentAura[5] and CurrentAura[4] and CurrentAura[4] > 0 then
										handler.CurrentSlot.Cooldown:SetCooldown(CurrentAura[5] - CurrentAura[4], CurrentAura[4])
										handler.CurrentSlot.Cooldown:Show()
									else
										handler.CurrentSlot.Cooldown:Hide()
									end
									
									handler.CurrentSlotNum = handler.CurrentSlotNum + 1
									
									handler.CurrentSlot:Show()
								else
									handler.BeginAtSlot = handler.CurrentSlotNum + 1
									break
								end
								
								CurrentAuraIndex = CurrentAuraIndex + 1
							end
						end
					end
				end
				
				if handler.CurrentSlotNum < 1 then
					for i=1, handler.AuraSlots do
						if handler.AuraHolder.AuraSlot[i]:IsVisible() then
							handler.AuraHolder.AuraSlot[i]:Hide()
						end
					end
				else
					for i=handler.CurrentSlotNum + 1, handler.EndAtSlot do
						if handler.AuraHolder.AuraSlot[i]:IsVisible() then
							handler.AuraHolder.AuraSlot[i]:Hide()
						end
					end
				end
				
				if handler.AuraGrowthDirectionX == "CENTER" then
					self:UpdateCenterPositioning(handler)
				end
			end
		end
end

function UAUR:UpdateAuras(Unit)
	if not UnitExists(Unit) then return end
	
	self:UpdateAuraCache(Unit)
	-- Update icons OnEvent instead of OnUpdate
	self:UpdateIcons(Unit)
end

-- Blizz already provides a list of possible colors
function UAUR:GetAuraColor(DType, Unit, AuraType)
	if AuraType == E.STR.HARMFUL then
		if DType then
			return DebuffTypeColor[DType]
		else
			return DebuffTypeColor["none"]
		end
	else
		return E:GetUnitReactionColor(Unit)
	end
end

function UAUR:__Construct()
	for k, v in pairs(UF.Frames) do
		if v.RawUnit ~= "player" then
			self:BuildAuras(v) -- Build aura frames for Unitframes
		end
	end
end

function UAUR:Init()
	--self.db = CO.db.auras
	
	self:__Construct()
end

E:AddModule("UnitAuras", UAUR)